/*
 * Echo Adaption Bridge Layer for CEM-PCI Card
 *
 * Copyright (C) 2011 Embedded Dynamics Inc.
 * All rights reserved.
 *
 */


#include <linux/init.h>

#include <linux/kernel.h>	/* printk() */
#include <linux/errno.h>	/* error codes */
#include <linux/module.h>
#include <linux/types.h>	/* size_t */
#include <linux/fcntl.h>	/* O_ACCMODE */
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/pci.h>		/* for PCI structures */
#include <linux/delay.h>
#include <asm/io.h>
#include <linux/spinlock.h>
#include <linux/device.h>	/* dev_err() */
#include <linux/interrupt.h>
#include <asm/system.h>		/* cli(), *_flags */
#include <asm/uaccess.h>	/* copy_*_user */
#include <linux/workqueue.h>	/* work_struct */
#include <linux/timer.h>	/* timer_struct */
#include <linux/moduleparam.h>
#include <linux/proc_fs.h>

#include <dahdi/kernel.h>

#include "wcb4xxp.h"

#include "cem_ec_bridge.h"
#include "ec_zlxxxxx.h"


#define CHANNELS_32	1

#define EC_MODULE_NAME          "zarlink_LEC_chip: "
#define DEBUG(args...)          printk(KERN_DEBUG EC_MODULE_NAME args);
#define ERROR(args...)          printk(KERN_ERR EC_MODULE_NAME args);
#define INFO(args...)           printk(KERN_INFO EC_MODULE_NAME args);

/*************************** Globals Variables  *******************************/
#define CREATE_LEC_PROCFS_ENTRY 
#ifdef CREATE_LEC_PROCFS_ENTRY
#define PROCFS_NAME 		"lec"
static struct proc_dir_entry *myproc;
#endif


static int first_init = 1;
//Debug mede
static int debug = 0; //By default DEBUG is off/DEBUG

// Our variables related to echo canceler and set to defaults.
static EC_STAT ec_stat = EC_NA;

// 16 is for ZL38065. 8 is for ZL50235. It will be set at init while checking
// the type of EC and will further indicate this type.
static int cfgNumChanGroups = 0;
static int cfgNumChan = 0;

//Default encoding should be set to A-Law
static int encoding = 2;

//Double tail should be set to no by default (PR1-Appliance)
static int doubletail = 0;

static void cem_echocan_free(struct dahdi_chan *chan, struct dahdi_echocan_state *ec);

static const struct dahdi_echocan_features cem_ec_features = {
	.NLP_automatic = 1,
	.CED_tx_detect = 1,
	.CED_rx_detect = 1,
};


static const struct dahdi_echocan_ops cem_ec_ops = {
	.echocan_free = cem_echocan_free,
};



/*************************** Protypes Functions *******************************/

void b4xxp_setreg8(struct b4xxp *b4, const unsigned int reg, const unsigned char val);
unsigned char b4xxp_getreg8(struct b4xxp *b4, const unsigned int reg);

/*************************** XHFC PCI Bridge Functions *******************************/
/*
 * HFC SRAM interface code.
 *  - Enable the auxiliary bridge and auxiliary bridge data lines
 *  V_BRG_EN | V_BRG_MD
 */
static inline void enablepcibridge(struct b4xxp *b4)
{
	b4xxp_setreg8(b4, R_BRG_PCM_CFG, 0x03);
	flush_pci();
}

static inline void disablepcibridge(struct b4xxp *b4)
{
	b4xxp_setreg8(b4, R_BRG_PCM_CFG, 0x02);
	flush_pci();
}

/* NOTE: read/writepcibridge do not use __pci_in/out because they are using b4->ioaddr not b4->addr */
static inline unsigned char readpcibridge(struct b4xxp *b4, unsigned char address)
{
	unsigned short cipv;
	unsigned char data;


/* slow down a PCI read access by 1 PCI clock cycle */
	b4xxp_setreg8(b4, R_CTRL, 0x4);
	wmb();

	if (address == 0)
		cipv=0x4000;
	else
		cipv=0x5800;

/* select local bridge port address by writing to CIP port */
	iowrite16(cipv, b4->ioaddr + 4);
	wmb();
	data = ioread8(b4->ioaddr);

/* restore R_CTRL for normal PCI read cycle speed */
	b4xxp_setreg8(b4, R_CTRL, 0x0);
	wmb();
	flush_pci();

	return data;
}


static inline void cpld_select(struct b4xxp *b4, unsigned char val)
{

	if ( val ) {
		b4xxp_setreg8(b4, R_GPIO_OUT1, 0x4);
	} else {
		b4xxp_setreg8(b4, R_GPIO_OUT1, 0);
	}
	flush_pci();
}

static inline int cpld_status(struct b4xxp *b4)
{
	return (b4xxp_getreg8(b4, R_GPIO_IN1) & 0x4) ? 1 : 0;
}

static inline int IsD7Set(struct b4xxp *b4)
{
	return (b4xxp_getreg8(b4, R_GPIO_IN1) & 0x8) ? 1 : 0;
}

static inline void writepcibridge(struct b4xxp *b4, unsigned char address, unsigned char data)
{
	unsigned short cipv;
	unsigned int datav;


	if (address == 0)
		cipv=0x4000;
	else
		cipv=0x5800;

/* select local bridge port address by writing to CIP port */
	iowrite16(cipv, b4->ioaddr + 4);
	wmb();

/* define a 32 bit dword with 4 identical bytes for write sequence */
	datav = data | ( (__u32) data <<8) | ( (__u32) data <<16) | ( (__u32) data <<24);

/*
 * write this 32 bit dword to the bridge data port
 * this will initiate a write sequence of up to 4 writes to the same address on the local bus
 * interface
 * the number of write accesses is undefined but >=1 and depends on the next PCI transaction
 * during write sequence on the local bus
 */
	iowrite32(datav, b4->ioaddr);
	wmb();
	flush_pci();
}

void __writeval(struct b4xxp *b4, u8 val)
{
	if ( val & 0x80)
		cpld_select(b4, 1);
	else
		cpld_select(b4, 0);

	writepcibridge(b4, 0, val);
}

void __writereg(struct b4xxp *b4, u8 val)
{
	if ( val & 0x80)
		cpld_select(b4, 1);
	else
		cpld_select(b4, 0);

	writepcibridge(b4, 1, val);
}


// Integred changes
void write_cpld(struct b4xxp *b4,  u8 reg, u8 val)
{	
        unsigned long flags;
	spin_lock_irqsave(&b4->seqlock, flags);
	enablepcibridge(b4);
	
	__writereg(b4, reg);
	__writeval(b4, val);
	
	disablepcibridge(b4);
	spin_unlock_irqrestore(&b4->seqlock, flags);
	
}

// Integred chnages.
int read_cpld(struct b4xxp *b4, u8 reg)
{
	u8 val=0;
        unsigned long flags;
	spin_lock_irqsave(&b4->seqlock, flags);
	enablepcibridge(b4);

	__writereg(b4, reg);
	val = readpcibridge(b4,0); 
	disablepcibridge(b4);
	spin_unlock_irqrestore(&b4->seqlock, flags);
	return val;
}

/** try reading Page 3 0x0c register, ECA register, checking out P37 of zl38065 DS
  as the zl38065 using 13 address length for its address, the page address located A12, A11,
  so, we need writing address 0x18 0x0c to sram bus, in more details
    0x18 = 1   1   0   0  0
        A12 A11 A10 A9 A8
    0x0c = 0000 1100
        A7-- ---A0
   thus, this function addr_l = 0x0c, addr_h = 0x18, val is return values of zl38065
   this function back is no D7 append, please append that with GPIO11 status
*/

int __read_reg(struct b4xxp *b4, u8 page, u8 reg)
{
	int val = 0;
        unsigned long flags;
        spin_lock_irqsave(&b4->seqlock, flags);
	enablepcibridge(b4);


	cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x08);  // Select ZL38065  
        cpld_select(b4, 0);
        writepcibridge(b4, 0, 0x00);  //CS_n=0
	
	cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x0a);
        
	// Select the register
	// Address range A7-A0 (addr_l )
	__writeval(b4, reg);


	cpld_select(b4, 0);
	writepcibridge(b4, 1, 0x0b);

	cpld_select(b4, 0);
	writepcibridge(b4, 0, page); //addr_h (A12..A8)

	cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x0c);
        cpld_select(b4, 0);
        writepcibridge(b4, 0, 0x01); // RWn=1 read mode
 
        cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x07);
        cpld_select(b4, 0);
        writepcibridge(b4, 0, 0x00); // DS_n=0
 
        cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x09);
        val = readpcibridge(b4,0);  //return data
	// Implement GPIO11 status
	if ( IsD7Set(b4) )
		val |= 0x80;

	// Reset 
        cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x07);
        cpld_select(b4, 0);
        writepcibridge(b4, 0, 0x1);  // DS_n=1
 
        cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x0c);
        cpld_select(b4, 0);
        writepcibridge(b4, 0, 0x0); //RWn=0
 
        cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x8);
        cpld_select(b4, 0);
        writepcibridge(b4, 0, 0x1); //CS_n=1
 
	disablepcibridge(b4);
	spin_unlock_irqrestore(&b4->seqlock, flags);
	return val;
}



void __write_reg(struct b4xxp *b4, u8 page, u8 reg, u8 val )
{
	
        unsigned long flags;
        spin_lock_irqsave(&b4->seqlock, flags);
	enablepcibridge(b4);

	cpld_select(b4, 0);
	writepcibridge(b4, 1, 0x08);
	cpld_select(b4, 0);
	writepcibridge(b4, 0, 0x00); //CS_n=0

	cpld_select(b4, 0);
	writepcibridge(b4, 1, 0x0a);   
	
	// Select the register
	// Address range A7-A0 (addr_l )
	__writeval(b4, reg);

	cpld_select(b4, 0);
	writepcibridge(b4, 1, 0x0b);
	cpld_select(b4, 0);
	writepcibridge(b4, 0, page);//addr_h (A12..A8)

	cpld_select(b4, 0);
	writepcibridge(b4, 1, 0x0c);
	cpld_select(b4, 0);
	writepcibridge(b4, 0, 0x00); // RWn=0 write mode

	cpld_select(b4, 0);
	writepcibridge(b4, 1, 0x09); 
	//write data to zl38065
	__writeval(b4, val);

        cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x07);
        cpld_select(b4, 0);
        writepcibridge(b4, 0, 0x00); //DS_n=0
 
        cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x07);
        cpld_select(b4, 0);
        writepcibridge(b4, 0, 0x01); //DS_n=1
 
        cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x0c);
        cpld_select(b4, 0);
        writepcibridge(b4, 0, 0x01);//RWn=0
 
        cpld_select(b4, 0);
        writepcibridge(b4, 1, 0x08);
        cpld_select(b4, 0);
        writepcibridge(b4, 0, 0x01); //CS_n=1
	disablepcibridge(b4);
	spin_unlock_irqrestore(&b4->seqlock, flags);

}
static u8 readreg(struct b4xxp *b4, u16 reg)
{
	u8 val = __read_reg(b4, (reg & 0xFF00) >> 8 , reg & 0xFF);
	if (debug)
		DEBUG("%s - page: %02x reg: %02x, val: %02x\n", __FUNCTION__, (reg & 0xFF00) >> 8 , reg & 0xFF, val);
	return val; 
}
static void writereg(struct b4xxp *b4, u16 reg, u8 val)
{
	if (debug)
		DEBUG("%s - page: %02x reg: %02x, val: %02x\n", __FUNCTION__, (reg & 0xFF00) >> 8 , reg & 0xFF, val);
	__write_reg(b4, (reg & 0xFF00) >> 8 , reg & 0xFF, val);
	
}

/*************************** End of XHFC PCI Bridge Functions *******************************/

const char *zarlink_echocan_name(const struct dahdi_chan *chan)
{
        struct b4xxp_span *bspan = container_of(chan->span, struct b4xxp_span,
                                                span);
        if (bspan->parent->card_type == CEMBRI)
                return "ZARLINK CEM";
        return NULL;
}

ECRETCODE_t zarlink_setencoding(struct b4xxp *b4 ,ECLAW_t code)
{
        u16 regv;    //* Byte ?
        int i;

        switch(code) {
        case LINEAR:
                for (i = 0; i < cfgNumChanGroups; i++ ) {
			
                        regv = readreg( b4, EC_MAIN_CTRL_REG_0 + i );
                        regv = regv & (~EC_Format);     //* Clear bits.
                        writereg(b4, EC_MAIN_CTRL_REG_0 + i, regv);        //* Write value.
                }
                if (debug)
                        DEBUG("Encoding updated to LINEAR.\n");
                break;
        case ULAW:
                for (i = 0; i < cfgNumChanGroups; i++ ) {
                        regv = readreg(b4,ECREGADDR(EC_MAIN_CTRL_REG_0 + i));
                        regv = regv & (~EC_Law_A);      //* Clear bits.
                        regv = regv | (EC_Format | EC_Law_u);           //* Set appropriate.
                        writereg(b4, EC_MAIN_CTRL_REG_0 + i,regv);        //* Write value.
                }
                if (debug)
                        DEBUG("Encoding updated to u-Law.\n");
                break;
        case ALAW:
                for (i = 0; i < cfgNumChanGroups; i++ ) {
                        regv = readreg( b4, EC_MAIN_CTRL_REG_0 + i );
                        regv = regv | (EC_Format | EC_Law_A);           //* Set appropriate.
                        writereg(b4, EC_MAIN_CTRL_REG_0 + i,regv);        //* Write value.
                }
                if (debug)
                        DEBUG("Encoding updated to a-Law.\n");
                break;
        default:
                ERROR("invalid encoding %d set.\n", code);
                return FAILURE;
        }
        return SUCCESS;
}



void zarlink_softreset(struct b4xxp *b4){

        int i;
	

        //* Make a sw reset
        for (i = 0; i < cfgNumChanGroups; i++ ) {
                writereg(b4,  EC_MAIN_CTRL_REG_0 + i,EC_PWUP);
        }
        //* Here should wait 250us
        udelay(250/*us*/);
        debug=1;
	printk("TESTING : %02x\n", readreg(b4, 0x8));
	printk("TESTING : %02x\n", readreg(b4, 0x9));
	printk("TESTING : %02x\n", readreg(b4, 0x1D));
	debug=0;
	//* Next disable all.
        for (i = 0; i < cfgNumChanGroups; i++ ) {
                writereg(b4, EC_MAIN_CTRL_REG_0 + i,0);
        }
        //* And here should wait 500us for PLL to relock.
        udelay(500/*us*/);
        //* Next disable interrupts and enable Rout & Sout. 
        writereg(b4, EC_MAIN_CTRL_REG_0, EC_ODE | EC_MIRQ | EC_MTDBI | EC_MTDAI | EC_Law_A  | EC_PWUP | EC_Format);
        for (i = 1; i < cfgNumChanGroups; i++ ) {
                writereg(b4, EC_MAIN_CTRL_REG_0+i, EC_MTDBI | EC_MTDAI | EC_Law_A | EC_PWUP | EC_Format);
        }
        //* Wait another 250us for ECA & ECB to execute their init procedures.
        udelay(250/*us*/);

        zarlink_setencoding(b4, encoding);
}

/****************************************************************************/
// Functions to export.
//  Case 64ms - echo celceller alocation for the TDM (here 0, 1 ... are groups)
//      |       |       |       |       |       
//       echo0A  echo0B  echo1A  echo1B    ....
//
//  Case 128ms - echo celceller alocation for the TDM (here 0, 1 ... are groups)
//      |               |               |               |             |       
//        echo0A+echo0B      HiZ          echo1A+echo1B     HiZ
//
//  
//
ECRETCODE_t zarlink_setechomode(struct b4xxp *b4,int channel, int tail_ms)
{
        unsigned short regv;    //* Byte ?
        unsigned int offs;
        int chGroup = channel/2;
        int chAB = channel%2;
        unsigned long flags;
        spinlock_t lock;
	 spin_lock_init(&lock);

        if (debug)
                DEBUG("Echo canceller at channel %d, tail set to %d [ms]\n", channel, tail_ms);

        if (channel < 0 || channel >= 2*cfgNumChanGroups) {
                ERROR("invalid channel number %d\n", channel);
                return FAILURE;
        }
        switch(tail_ms) {
        case 0:

                //* If we would like to bypass
                spin_lock_irqsave(&lock, flags);
                offs = chGroup * EC_GROUP_OFFSET;
                regv = readreg(b4, (offs + ECA_OFFSET + EC_CONTROL_REG_1)) ;
                writereg(b4,      //* Clear  ExtDelay in echo A
                        (offs + ECA_OFFSET + EC_CONTROL_REG_1), regv & (~EC_ExtDl));
                regv = readreg( b4, ((offs + chAB * ECB_OFFSET + EC_CONTROL_REG_1)) );
                //udelay(1); //dpn: In order to correctly write to Control Register 1 and 2 of ECB, 
                writereg(b4,      //* Set Bypass of the appropriate echo
                        (offs + chAB * ECB_OFFSET + EC_CONTROL_REG_1),regv | EC_Bypass);
                spin_unlock_irqrestore(&lock, flags);
                break;
        case 64:
                //* If normal delay is required.
                spin_lock_irqsave(&lock, flags);
                offs = chGroup * EC_GROUP_OFFSET;
                regv = readreg( b4, (offs + ECA_OFFSET + EC_CONTROL_REG_1) );
                writereg(b4,    //* Clear ExtDelay in echo A
                        (offs + ECA_OFFSET + EC_CONTROL_REG_1), regv & (~EC_ExtDl));
                regv = readreg(b4, ((offs + chAB * ECB_OFFSET + EC_CONTROL_REG_1)) );
                writereg(b4, //* Clear Bypass of the appropriate echo 
                        (offs + chAB * ECB_OFFSET + EC_CONTROL_REG_1),regv & (~EC_Bypass));
                udelay(1); //dpn: In order to correctly write to Control Register 1 and 2 of ECB, 
                           //it is necessary to write the data twice to the register, one immediately 
                           //after another. The two writes must be separated by at least 350 ns and 
                           //no more than 20 us.
                writereg(b4,     //* Clear Bypass of the appropriate echo 
                        (offs + chAB * ECB_OFFSET + EC_CONTROL_REG_1), regv & (~EC_Bypass));
                spin_unlock_irqrestore(&lock, flags);
                break;
#ifndef CHANNELS_8
        case 128:
        if (doubletail) {
                //* If extended delay is required init both A & B.
                spin_lock_irqsave(&lock, flags);
                offs = chGroup * EC_GROUP_OFFSET;
                regv = readreg( b4, (offs + ECA_OFFSET + EC_CONTROL_REG_1) );
                regv = regv | EC_ExtDl;         //Set  ExtDelay in echo A       
                regv = regv & (~EC_Bypass);     //Clear bypass of A echo
                writereg(b4, (offs + ECA_OFFSET + EC_CONTROL_REG_1), regv);
                regv = readreg( b4, ((offs + ECB_OFFSET + EC_CONTROL_REG_1)) );
                writereg(b4, (offs + ECB_OFFSET + EC_CONTROL_REG_1),regv & (~EC_Bypass)); //Clear bypass of echo B
                udelay(1); //dpn: In order to correctly write to Control Register 1 and 2 of ECB, 
                //it is necessary to write the data twice to the register, one immediately 
                //after another. The two writes must be separated by at least 350 ns and 
                //no more than 20 us.
                writereg(b4, (offs + ECB_OFFSET + EC_CONTROL_REG_1),regv & (~EC_Bypass)); //Clear bypass of echo B
                spin_unlock_irqrestore(&lock, flags);
        }
        break;
#endif
        default:
                ERROR("invalid tail(ms) set %d.\nWe will run in bypass mode\n", tail_ms);
                //* If we would like to bypass
                spin_lock_irqsave(&lock, flags);
                offs = chGroup * EC_GROUP_OFFSET;
                regv = readreg( b4 , (offs + ECA_OFFSET + EC_CONTROL_REG_1)) ;
                writereg(b4,      //* Clear  ExtDelay in echo A
                        (offs + ECA_OFFSET + EC_CONTROL_REG_1), regv & (~EC_ExtDl));
                regv = readreg( b4, (offs + chAB * ECB_OFFSET + EC_CONTROL_REG_1) );
                //udelay(1); //dpn: In order to correctly write to Control Register 1 and 2 of ECB, 
                writereg(b4,     //* Set Bypass of the appropriate echo
                        (offs + chAB * ECB_OFFSET + EC_CONTROL_REG_1), regv | EC_Bypass);
                spin_unlock_irqrestore(&lock, flags);
                break;
        }

        return SUCCESS;
}


/********************************************************************************* proc stuff *****/

#ifdef CREATE_LEC_PROCFS_ENTRY

typedef unsigned char zl_reg_type;

#define COUNT_OF_CTRL_ROW	6


enum ec_status_row_type{
	EC_ST_ROW_CHANNEL=0,
	EC_ST_ROW_BYPASS,
	EC_ST_ROW_ACTIVE,
	EC_ST_ROW_NB,
	EC_ST_ROW_TD,
	EC_ST_ROW_TDG,
	EC_ST_ROW_DTDET,
	EC_ST_ROW_RPEAK,
	EC_ST_ROW_SPEAK,
	COUNT_OF_STATUS_ROW
};



zl_reg_type get_reg(struct b4xxp *b4, int channel, int reg) {
	int chB = channel%2;
	int ch_group = channel/2;
	int group_offs = ch_group * EC_GROUP_OFFSET;
	zl_reg_type regv;
	int adrs = reg + group_offs + chB*ECB_OFFSET;
	regv = readreg( b4, adrs );
	return regv;
}
#define TEST_FLAG(reg, flag)	(reg & flag)
int strlcat_status_table(struct b4xxp *b4, char *ptable, size_t len_of_table)
{
	char *first_col[COUNT_OF_STATUS_ROW] = {
				"Channels                 ", 
				"Bypass the EC            ", 
				"Active (Rin>LP threshold)",
				"Narrow-band signal on Rin",
				"2100Hz tone              ",
				"2100Hz tone bypass the EC",
				"Double-talk              ",
				"Rin Peak %               ",
				"Sin Peak %               ",
				};
	int i;
	u16 signal_level;
	int ch;
	char temp[20];
	zl_reg_type flags_status_reg[] = { EC_ACTIVE, EC_NB, EC_TD, EC_TDG, EC_DTDet };
	zl_reg_type regv;
	for (i = EC_ST_ROW_CHANNEL; i < COUNT_OF_STATUS_ROW; i++){
		strlcat( ptable, first_col[i], len_of_table);
		switch(i){
			case EC_ST_ROW_CHANNEL:
			for (ch = 0; ch < cfgNumChan; ch++ ) {
				snprintf( temp, sizeof(temp), " %02u", ch+1);
				strlcat( ptable, temp, len_of_table);
			}
			strlcat( ptable, "\n", len_of_table);
			break;

			case EC_ST_ROW_BYPASS:
			for (ch = 0; ch < cfgNumChan; ch++ ) {
				regv = get_reg( b4, ch, EC_CONTROL_REG_1 );
				if( TEST_FLAG(regv, EC_Bypass) ){
					strlcat( ptable, "  *", len_of_table);
				}else{
					strlcat( ptable, "  -", len_of_table);
				}
			}
			strlcat( ptable, "\n", len_of_table);
			break;

			case EC_ST_ROW_ACTIVE:
			case EC_ST_ROW_NB:
			case EC_ST_ROW_TD:
			case EC_ST_ROW_TDG:
			case EC_ST_ROW_DTDET:
			for (ch = 0; ch < cfgNumChan; ch++ ) {
				regv = get_reg(b4, ch, EC_CONTROL_REG_1 );
				if( TEST_FLAG(regv, EC_Bypass) ){
					strlcat( ptable, "  ?", len_of_table);
				}else{
					regv = get_reg(b4, ch, EC_STATUS_REG );
					if( TEST_FLAG(regv, flags_status_reg[i-EC_ST_ROW_ACTIVE]) ){
						strlcat( ptable, "  *", len_of_table);
					}else{
						strlcat( ptable, "  -", len_of_table);
					}
				}
			}
			strlcat( ptable, "\n", len_of_table);
			break;

			case EC_ST_ROW_RPEAK:
			case EC_ST_ROW_SPEAK:
			for (ch = 0; ch < cfgNumChan; ch++ ) {
				regv = get_reg(b4, ch, EC_CONTROL_REG_1 );
				if( TEST_FLAG(regv, EC_Bypass) ){
					strlcat( ptable, "  ?", len_of_table);
				}else{
					if( i == EC_ST_ROW_RPEAK ){
						signal_level = get_reg( b4, ch, EC_RIN_PEAK_DETECT_REG_L );
						signal_level |= (u16)get_reg(b4, ch, EC_RIN_PEAK_DETECT_REG_H )<<8;
					}else{
						signal_level = get_reg(b4, ch, EC_SIN_PEAK_DETECT_REG_L );
						signal_level |= (u16)get_reg(b4, ch, EC_SIN_PEAK_DETECT_REG_H )<<8;
					}
					snprintf( temp, sizeof(temp), " %02u", ((signal_level>>6)*100)/255 );
					strlcat( ptable, temp, len_of_table);
				}
			}
			strlcat( ptable, "\n", len_of_table);
			break;
		}
	}
	strlcat( ptable, "                          Legend: (*)Active flag (-)Not active flag (?)Unknown value\n", len_of_table);
	return strlen(ptable);
}


int strlcat_ctrl_reg_table(struct b4xxp *b4, char *ptable, size_t len_of_table)
{

	char *first_col[COUNT_OF_CTRL_ROW] = {	
				"Channels                 ", 
				"CONTROL_REG1             ",
				"CONTROL_REG2             ",
				"CONTROL_REG3             ",
				"CONTROL_REG4             ",
				"STATUS                   "
				};
	int i;
	int ch;
	char temp[20];
	int reg_table[] = { EC_CONTROL_REG_1, EC_CONTROL_REG_2, EC_CONTROL_REG_3, EC_CONTROL_REG_4, EC_STATUS_REG };
	zl_reg_type regv;
	for (i = 0; i < COUNT_OF_CTRL_ROW; i++){
		strlcat( ptable, first_col[i], len_of_table);
		switch(i){
			case 0:
			for (ch = 0; ch < cfgNumChan; ch++ ) {
				snprintf( temp, sizeof(temp), " %02u", ch+1);
				strlcat( ptable, temp, len_of_table);
			}
			strlcat( ptable, "\n", len_of_table);
			break;

			case 1:
			case 2:
			case 3:
			case 4:
			case 5:
			for (ch = 0; ch < cfgNumChan; ch++ ) {
				regv = get_reg(b4, ch, reg_table[i-1] );
				snprintf( temp, sizeof(temp), " %02X", regv);
				strlcat( ptable, temp, len_of_table);
			}
			strlcat( ptable, "\n", len_of_table);
			break;
		}
	}
	return strlen(ptable);
}


static int lec_proc_read(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	struct b4xxp *b4_cards = data;
	int len, res;
	len = sprintf(buf,
                      "Number of available channels at 64 ms..: %d\n"
                      "Encoding is set to.....................: %s\n"
                      "Maximum tail length....................: %d ms\n",
			cfgNumChanGroups * 2, encoding == 2 ? "a-law":"u-law", 
#ifndef CHANNELS_8
			doubletail ? 128:64
#else
			64
#endif
                      );

	res = strlcat_status_table(b4_cards,buf, count);
	if( res == 0 ){
		strlcat( buf, "ERROR:Can not read status", count);
	}
	if( debug ){
		res = strlcat_ctrl_reg_table(b4_cards, buf, count);
		if( res == 0 ){
			strlcat( buf, "ERROR:Can not read control register", count);
		}
	}
	 *eof=1;
	len = strlen( buf );
        return len;
}


#endif /* CREATE_WCB4XXP_PROCFS_ENTRY */




static void cem_echocan_free(struct dahdi_chan *chan, struct dahdi_echocan_state *ec)
{
	struct b4xxp_span *bspan = container_of(chan->span, struct b4xxp_span, span);
	int channel;

	memset(ec, 0, sizeof(*ec));

	if (debug)
		printk("Disabling echo cancellation on chan %d span %d\n", chan->chanpos, chan->span->offset);

	//channel = (chan->span->offset * 8) + ((chan->chanpos - 1) * 4) + 1;
	// Channel Channel 0 = Span 0 + 0B
	// Channel 1 = Span 0 + 1B = 
	// Channel 3 = Span 1 + 0B =  
	channel = chan->channo -1 ;

	//channel = (chan->span->offset * 8) + ((chan->chanpos - 1) * 4) + 1;

	//ec_write(bspan->parent, chan->chanpos - 1, channel, 0x01);

	zarlink_setechomode(bspan->parent, channel, 0);
}



int zarlink_echocan_create(struct dahdi_chan *chan,
				struct dahdi_echocanparams *ecp,
				struct dahdi_echocanparam *p,
				struct dahdi_echocan_state **ec)
{
	struct b4xxp_span *bspan = container_of(chan->span, struct b4xxp_span, span);
	int channel;
	printk("Inside zarlink_echocan_create\n");


	if ( zarlink_is_ec(bspan->parent) == 0 ) 
	{
		printk(KERN_WARNING "Couldn't Find E/C Module\n");
		return -ENODEV;
	}


	if (chan->chanpos == 3) {
		printk(KERN_WARNING "Cannot enable echo canceller on D channel of span %d; failing request\n", chan->span->offset);
		return -EINVAL;
	}

	*ec = &bspan->ec[chan->chanpos];
	(*ec)->ops = &cem_ec_ops;
	(*ec)->features = cem_ec_features;

	if (debug)
		printk("Enabling echo cancellation on chan %d span %d\n", chan->chanpos, chan->span->offset);

	//channel = (chan->span->offset * 8) + ((chan->chanpos - 1) * 4) + 1;
	//channel =1 + ((chan->chanpos ) +  chan->span->offset );
	channel = chan->channo -1 ;
	
	printk(" Creating channel : %d, tail %d \n", channel, ecp->tap_length);
	
	zarlink_setechomode(bspan->parent, channel, ecp->tap_length);

	return 0;
}

int zarlink_is_ec(struct b4xxp *b4)
{
	u8 gpio =  b4xxp_getreg8(b4, R_GPI_IN2);
        printk("THE GPIO is:%02x\n",gpio);

	if ( (gpio & 0x2) == 2 )
	{
		return 0;
	}
	return 1;

}

void zarlink_ec_init(struct b4xxp *b4)
{

	u8 gpio =  b4xxp_getreg8(b4, R_GPI_IN2);

        ec_stat = EC_NA;
        cfgNumChanGroups = 0;


	printk("Inside zarlink_CEM_ec_init\n");
	printk("Gpio: %02x\n", gpio);
	printk("EC_Insert: %d\n", gpio & 0x2);
	printk("EC_Type: %d\n", gpio & 0x4);


	if ( zarlink_is_ec(b4) == 0 )
		return;

	if ( (gpio & 0x4) == 0x4 )
	{
		printk(" Chip Type: zl38065/zl50232\n");
	} else {
		printk(" Chip Type: zl50234/zl50235\n");
	}

        ec_stat = EC_FREE;
#if CHANNELS_4
		cfgNumChanGroups = EC4_NUMOFGROUPS;
		INFO("Selected chip ZL50233 with 4 channels.\n");
#elif CHANNELS_8
		cfgNumChanGroups = EC8_NUMOFGROUPS;
		INFO("Selected chip ZL50234 with 8 channels \n");
#elif  CHANNELS_16
		cfgNumChanGroups = EC16_NUMOFGROUPS;
		INFO("Select chip ZL50235 with 16 channels\n");
#elif  CHANNELS_32
		cfgNumChanGroups = EC32_NUMOFGROUPS;
		INFO("Select chip ZL38065/ZL50232 with 32 channels\n");
#endif
	

        INFO("Inserting  EchoCanceler module. %d channels\n",2*cfgNumChanGroups);


        udelay(500);
        zarlink_softreset(b4); 

        if(debug)
                DEBUG("Initial encoding set to %d\n",encoding);


#ifdef CREATE_LEC_PROCFS_ENTRY
	if ( first_init )
	{
		if (!(myproc = create_proc_read_entry(PROCFS_NAME, 0444, NULL,
		                             lec_proc_read, b4))) {
			printk(KERN_ERR "%s: ERROR: Could not initialize /proc/%s\n",THIS_MODULE->name, PROCFS_NAME);
		}
		first_init = 0;
	}
#endif
//	for (i = 0; i < cfgNumChanGroups; i++ ) {
//		zarlink_setechomode(b4, i , 0);
//	}
cfgNumChan = cfgNumChanGroups * 2;
}



void zarlink_ec_exit(void)
{
#ifdef CREATE_LEC_PROCFS_ENTRY
	if (first_init == 0)
		remove_proc_entry(PROCFS_NAME, NULL);
#endif

}
